import yaml
import os
from os import path


class Configuration (object):
    def __init__(self, environ):
        home_path = environ['HOME']
        # per xdg spec:
        # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
        config_home = environ.get(
            'XDG_CONFIG_HOME', path.join(home_path, '.config'))
        data_home = environ.get(
            'XDG_DATA_HOME', path.join(home_path, '.local/share'))
        self.config_path = path.join(config_home, 'slipway.yaml')
        self.data_directory = path.join(data_home, 'slipway')
        self.runtime_dir = environ.get(
            'XDG_RUNTIME_DIR',
            path.join('/run/user', str(os.getuid())))
        self.workspace = path.join(home_path, 'workspace')
        self.alias = {}
        self.environment = []
        self.volume = []
        self.pull = False
        self.pull_daily = False
        self.mount_docker = False
        self.runtime = 'docker'
        self.network = 'host'

    def load(self):
        if path.exists(self.config_path):
            with open(self.config_path) as file_descriptor:
                config = yaml.safe_load(file_descriptor)

                if 'alias' in config:
                    self.alias = config['alias']
                if 'workspace' in config:
                    self.workspace = config['workspace']
                if 'environment' in config:
                    self.environment = config['environment']
                if 'volume' in config:
                    self.volume = config['volume']
                if 'pull' in config:
                    self.pull = config['pull']
                if 'pull_daily' in config:
                    self.pull_daily = config['pull_daily']
                if 'mount_docker' in config:
                    self.mount_docker = config['mount_docker']
                if 'runtime' in config:
                    self.runtime = config['runtime']



from .util import snake_case


class Volume(object):
    def __init__(self, image_name, path):
        base = 'slipway_' + snake_case(image_name)
        self.name = base + '_' + snake_case(path)
        self.path = path


class Volumes(object):
    def __init__(self, client, args, image):
        self.client = client
        self.args = args
        self.image = image

    def list(self):
        """
        Lists all volumes
        """
        return map(
            lambda volume: Volume(self.image.name, volume), self.image.volumes)

    def purge(self):
        """
        Clears all volumes tied to the given arguments
        """
        current_volumes = self.client.list_volume_names()
        for volume in self.list():
            found = None
            for volume_name in current_volumes:
                if volume.name == volume_name:
                    found = volume_name
            if found is not None:
                self.client.remove_volume(found)

    def initialize(self):
        """
        Checks if the required volume are present and creates them if missing.
        """
        current_volume_names = self.client.list_volume_names()
        for volume in self.list():
            exist = False
            for volume_name in current_volume_names:
                if volume.name == volume_name:
                    exist = True
            if not exist:
                self.client.create_volume(volume.name)



from os import environ
import sys
from argparse import ArgumentParser

from .container import Container
from .configuration import Configuration
from .client import PodmanClient, DockerClient


def parse_args(configuration):
    parser = ArgumentParser()
    parser.add_argument(
        '--data-directory',
        default=configuration.data_directory)
    subparsers = parser.add_subparsers(dest='mode')
    subparsers.required = True

    start_parser = subparsers.add_parser('start')
    start_parser.add_argument('image')
    start_parser.add_argument(
        '--mount-docker',
        action='store_true',
        default=configuration.mount_docker)
    start_parser.add_argument(
            '--pull', action='store_true', default=configuration.pull)
    start_parser.add_argument(
        '--pull-daily', action='store_true', default=configuration.pull_daily)
    start_parser.add_argument(
        '--volume', '-v', action='append', default=configuration.volume)
    start_parser.add_argument(
        '--environment',
        '-e',
        action='append',
        default=configuration.environment)
    start_parser.add_argument(
        '--runtime', '-r', default=configuration.runtime)
    start_parser.add_argument(
        '--workspace', default=configuration.workspace)
    start_parser.add_argument(
        '--network', default=configuration.network)
    start_parser.add_argument(
        '--runtime-dir',
        default=configuration.runtime_dir)

    purge_parser = subparsers.add_parser('purge')
    purge_parser.add_argument('image')

    args = parser.parse_args()

    if args.image in configuration.alias:
        alias = configuration.alias[args.image]
        if 'image' in alias:
            args.image = alias['image']
        else:
            args.image = alias

        if 'environment' in alias:
            args.environment = alias['environment']

        if 'volume' in alias:
            args.volume.extend(alias['volume'])

        if 'network' in alias:
            args.network = alias['network']

    return args


def main():
    if sys.version_info.major < 3:
        print('Python 2 is not supported')
        sys.exit(1)
    configuration = Configuration(environ)
    configuration.load()
    args = parse_args(configuration)
    client = None
    if args.runtime == 'podman':
        client = PodmanClient()
    else:
        client = DockerClient()
    container = Container(client, args)
    if args.mode == 'start':
        container.image.initialize()
        container.volumes.initialize()
        container.binds.initialize()
        container.xdg_open.start_server()
        if container.exists():
            print('Container {} already exists'.format(container.name))
            sys.exit(1)
        container.run()
    elif args.mode == 'purge':
        container.volumes.purge()




import os
from os import path
from .image import Image
from .volumes import Volumes
from .binds import Binds
from .util import snake_case
from .xdg_open import XdgOpen


class Container(object):

    def __init__(self, client, args):
        self.client = client
        self.args = args
        self.xdg_open = XdgOpen(path.join(args.runtime_dir, 'slipway'))
        self.image = Image(self.client, self.args)
        self.volumes = Volumes(self.client, self.args, self.image)
        self.binds = Binds(self.client, self.args, self.image)
        self.name = 'slipway_' + snake_case(self.args.image)

    def _entrypoint_script(self):
        module_dir = path.dirname(path.abspath(__file__))
        return path.abspath(
            path.join(module_dir, './entrypoint.py'))

    def _volumes_env(self):
        return ','.join(list(map(lambda vol: vol.path, self.volumes.list())))

    def _append_docker_gid(self, arguments):
        with open('/etc/group') as file_descriptor:
            content = file_descriptor.read()
            for line in content.split('\n'):
                parts = line.split(':')
                if parts[0] == 'docker':
                    arguments.append('-e')
                    arguments.append('SLIPWAY_DOCKER_GID={}'.format(parts[2]))
                    break

    def _run_arguments(self):
        arguments = []
        if self.args.runtime == 'podman':
            arguments.append('podman')
        else:
            arguments.append('docker')

        arguments.extend([
            'run',
            # Required for strace and other debugging tools to work.
            '--cap-add', 'SYS_PTRACE',
            '--rm',
            '-ti',
            '--detach-keys', 'ctrl-q,ctrl-q',
            '--name', self.name,
            '-e', 'GH_USER',
            '-e', 'GH_PASS',
            '-e', 'DISPLAY',
        ])
        arguments.append('--network={}'.format(self.args.network))

        is_rootless = self.client.is_rootless()

        if not is_rootless:
            arguments.extend([
                '--user', 'root',
                '-e', 'SLIPWAY_USER={}'.format(self.image.user),
                '-e', 'SLIPWAY_VOLUMES={}'.format(self._volumes_env()),
                '-v', self._entrypoint_script() + ':/slipway-entrypoint.py:ro',
                '--entrypoint', 'python3'
            ])

        if is_rootless:
            arguments.extend([
                '--userns=keep-id'
            ])

        arguments.extend([
            '-v', '{}:/usr/bin/xdg-open'.format(self.xdg_open.client_path),
            '-v', '{}/slipway:/run/slipway'.format(
                self.args.runtime_dir),
        ])

        if self.args.mount_docker:
            arguments.append('-v')
            arguments.append('/run/docker.sock:/run/docker.sock')
            self._append_docker_gid(arguments)

        for env in self.args.environment or []:
            arguments.append('-e')
            arguments.append(env)

        for bind in self.binds.list():
            arguments.append('-v')
            argument = bind.host_path + ':' + bind.container_path
            if 'ro' in bind.type:
                argument += ':ro'
            arguments.append(argument)

        for volume in self.volumes.list():
            arguments.append('--mount')
            arguments.append(
                'type=volume,source={},target={}'
                .format(volume.name, volume.path))

        arguments.append(self.image.name)
        if not is_rootless:
            arguments.append('/slipway-entrypoint.py')
        arguments.append('tmux')
        arguments.append('new')

        return arguments

    def exists(self):
        return self.name in self.client.list_all_containers()

    def run(self):
        os.execvp(self.args.runtime, self._run_arguments())



from subprocess import Popen, DEVNULL
import os
from os import path


class XdgOpen(object):
    def __init__(self, slipway_dir):
        self._script_dir = path.dirname(path.realpath(__file__))
        self._slipway_dir = slipway_dir
        self._socket_file = path.join(slipway_dir, 'xdg-open.sock')
        self._pid_file = path.join(slipway_dir, 'xdg-open.pid')

    @property
    def runtime_dir(self):
        return self._slipway_dir

    @property
    def client_path(self):
        return path.join(self._script_dir, 'client.py')

    def _kill(self, code):
        with open(self._pid_file) as file:
            pid = str(file.read())
            os.kill(int(pid), code)

    def is_server_running(self):
        try:
            # per documentation, this will just check if the process is
            # running:
            # https://linux.die.net/man/2/kill
            self._kill(0)
            return True
        except OSError:
            return False

    def _remove_file(self, file):
        try:
            os.remove(file)
        except FileNotFoundError:
            pass

    def _start_background_process(self, child_process_args):
        script_dir = path.dirname(path.realpath(__file__))
        server_script = path.join(script_dir, 'server.py')
        process = Popen(
            ['python3', server_script, self._socket_file],
            stdin=DEVNULL,
            stdout=DEVNULL,
            stderr=DEVNULL,
            **child_process_args)
        with open(self._pid_file, 'w+') as file:
            file.write(str(process.pid))

    def start_server(self, **kwargs):
        if not path.exists(self._slipway_dir):
            os.makedirs(self._slipway_dir)

        if path.exists(self._pid_file):
            if not self.is_server_running():
                self._remove_file(self._pid_file)
                self._remove_file(self._socket_file)
                self._start_background_process(kwargs)
        else:
            self._start_background_process(kwargs)

    def stop_server(self):
        if self.is_server_running():
            self._kill(15)  # SIGTERM



import pwd
import os
from sys import platform
from os import environ, path


class Bind(object):
    def __init__(self, host_path, container_path, type='d'):
        self.host_path = host_path
        self.container_path = container_path
        self.type = type


class Binds(object):
    """
    A bind is a type of volume which maps a directory or file from the host
    into the container.
    """

    def __init__(self, client, args, image):
        self.client = client
        self.args = args
        self.image = image

    def initialize(self):
        """
        Creates the directory with appropriate permissions if it doesn't
        exist. Otherwise if we let docker do this the permissions will be
        set as root, which we don't want.
        """
        for bind in self.list():
            if not path.exists(bind.host_path):
                if 'd' in bind.type:
                    os.makedirs(bind.host_path)
                elif 'f' in bind.type:
                    dirname = path.dirname(bind.host_path)
                    if not path.exists(dirname):
                        os.makedirs(path.dirname(bind.host_path))
                    open(bind.host_path, 'w+').close()

    def _gpg_binds(self):
        uid = pwd.getpwnam(environ['USER']).pw_uid
        gnupg_socket_path = path.join('/run/user', str(uid), 'gnupg')
        if path.exists(gnupg_socket_path):
            # FIXME: how to get uid of user in container?
            yield Bind(gnupg_socket_path, '/run/user/1000/gnupg', 'd')

        gnupg_path = path.join(environ['HOME'], '.gnupg')
        if path.exists(gnupg_path):
            container_path = path.join(self.image.home, '.gnupg')
            yield Bind(gnupg_path, container_path, 'd')

    def list(self):
        """
        List all bind mounts, including the default ones.
        """
        home_mappings = [
            ('d', '.ssh'),
            ('f', '.gitconfig'),
            # TODO: find a way to delegate to the image instead.
            ('f', '.yarnrc'),
            ('f', '.cargo/credentials'),
            ('f', '.pypirc')
        ]
        for type, mapping in home_mappings:
            host_path = path.join(environ['HOME'], mapping)
            container_path = path.join(self.image.home, mapping)
            yield Bind(host_path, container_path, type)

        workspace_base = path.basename(self.args.workspace)
        workspace_path = path.join(self.image.home, workspace_base)
        yield Bind(self.args.workspace, workspace_path, 'd')

        if path.isfile('/etc/localtime') and platform != 'darwin':
            yield Bind('/etc/localtime', '/etc/localtime', 'fro')
        if path.exists('/tmp/.X11-unix'):
            yield Bind('/tmp/.X11-unix', '/tmp/.X11-unix', 'fro')

        yield from self._gpg_binds()

        for volume in self.args.volume or []:
            parts = volume.split(':')
            host_path = parts[0]
            container_path = parts[1]
            type = 'd'
            if path.isfile(host_path):
                type = 'f'
            if len(parts) == 3 and parts[2] == 'ro':
                type += 'ro'
            yield Bind(host_path, container_path, type)



import subprocess
import json


class PodmanClient(object):
    def list_volume_names(self):
        result = subprocess.run(
            ['podman', 'volumes', 'ls', '--format', '{{.Name}}'],
            capture_output=True
        )
        return str(result.stdout, 'utf8').strip().split('\n')

    def create_volume(self, name):
        subprocess.run(
            ['podman', 'volume', 'create', name],
            capture_output=True
        )

    def pull_image(self, name):
        subprocess.run(['podman', 'pull', name])

    def inspect_image(self, name):
        result = subprocess.run(
            ['podman', 'image', 'inspect', '--format', '{{json .}}', name],
            capture_output=True
        )
        images = json.loads(str(result.stdout, 'utf8').strip())
        return images[0]

    def remove_volume(self, name):
        result = subprocess.run(
            ['podman', 'volumes', 'remove', name])
        assert(result.returncode == 0)

    def image_exists(self, name):
        result = subprocess.run(
            ['podman', 'image', 'inspect', name],
            capture_output=True)
        return result.returncode == 0

    def is_rootless(self):
        result = subprocess.run(
            ['podman', 'info', '-f', '{{json .}}'],
            capture_output=True)
        info = json.loads(str(result.stdout, 'utf8').strip())

        return info['host']['rootless']

    def list_all_containers(self):
        result = subprocess.run(
            ['podman', 'ps', '-a', '--format', '{{.Names}}'],
            capture_output=True)
        return str(result.stdout, 'utf-8').strip().split('\n')




from docker.errors import ImageNotFound
import docker
import subprocess


class DockerClient(object):
    def __init__(self):
        self.client = docker.from_env()

    def list_volume_names(self):
        return map(lambda volume: volume.name, self.client.volumes.list())

    def create_volume(self, name):
        self.client.volumes.create(name)

    def remove_volume(self, name):
        self.client.volumes.get(name).remove()

    def pull_image(self, name):
        subprocess.run(['docker', 'pull', name])

    def image_exists(self, name):
        try:
            self.client.images.get(name)
            return True
        except ImageNotFound:
            return False

    def inspect_image(self, name):
        return self.client.images.get(name).attrs

    def is_rootless(self):
        # TODO
        return False

    def list_all_containers(self):
        return [
            container.name
            for container in self.client.containers.list(all=True)
        ]



#!/usr/bin/env python3

import socket
import json
import sys
from os import path, environ

url = sys.argv[1]

socket_file = path.join(
    environ.get('SLIPWAY_RUNTIME_DIR', '/run/slipway'),
    'xdg-open.sock')

with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
    sock.connect(socket_file)
    sock.sendall(bytes(url + '\n', 'utf8'))
    line_end = False
    attempts_left = 100
    data = ''

    while not line_end and attempts_left:
        attempts_left -= 1
        chunk = sock.recv(2046)
        converted = str(chunk, 'utf8')
        if '\n' in converted:
            line_end = True
        data += converted

    if attempts_left:
        result = json.loads(data)
        sys.stderr.write(result['stderr'])
        sys.stdout.write(result['stdout'])
        sys.exit(result['returncode'])
    else:
        print('xdg-open server failed to reply', file=sys.stderr)
        sys.exit(1)



from socketserver import UnixStreamServer, StreamRequestHandler
from subprocess import run, PIPE
import json
import sys
import os

socket_file = sys.argv[1]


class XdgOpenHandler(StreamRequestHandler):
    def handle(self):
        url = self.rfile.readline().strip()
        try:
            result = run(
                ['xdg-open', url],
                shell=False,
                check=False,
                stdout=PIPE,
                stderr=PIPE)
            response = json.dumps({
                'returncode': result.returncode,
                'stdout': str(result.stdout, 'utf8'),
                'stderr': str(result.stderr, 'utf8')
            })

            self.request.sendall(bytes(response + '\n', 'utf8'))
        except PermissionError:
            response = json.dumps({
                'returncode': 127,
                'stdout': '',
                'stderr': 'xdg-open not found on host machine'
            })
            self.request.sendall(bytes(response + '\n', 'utf8'))


with UnixStreamServer(socket_file, XdgOpenHandler) as server:
    server.serve_forever()




from os import path, makedirs
from datetime import datetime
from .util import snake_case


class Image(object):
    def __init__(self, client, args):
        self.client = client
        self.args = args
        self.name = args.image
        self._metadata = None

    def _image_metadata(self):
        if self._metadata is None:
            self._metadata = self.client.inspect_image(self.name)
        return self._metadata

    def pull(self):
        self.client.pull_image(self.name)

    def exists(self):
        """
        Returns true if the image exists locally, false otherwise.
        """

        return self.client.image_exists(self.name)

    def _pulled_today(self):
        """
        Returns true is the last successful pull was performed today.
        """
        last_stale_path = path.join(
            self.args.data_directory,
            'last_stale_check',
            snake_case(self.name))
        if path.exists(last_stale_path):
            with open(last_stale_path) as file_descriptor:
                content = file_descriptor.read()
                last_pull = datetime.strptime(content, '%Y-%m-%d')
                now = datetime.now()
                return now.day == last_pull.day \
                    and now.month == last_pull.month \
                    and now.year == last_pull.year

        return False

    def _create_stale_check_file(self):
        """
        Creates the last_pull file which after a successful pull was performed.
        """
        stale_check_dir = path.join(
            self.args.data_directory, 'last_stale_check')
        makedirs(stale_check_dir, exist_ok=True)
        content = datetime.now().strftime('%Y-%m-%d')
        last_stale_path = path.join(stale_check_dir, snake_case(self.name))
        with open(last_stale_path, 'w+') as file_descriptor:
            file_descriptor.write(content)

    def initialize(self):
        """
        Pulls the image if not present
        """
        if not self.exists():
            message = 'Image {} not found, attempting to pull down'
            print(message.format(self.name))
            self.pull()
            self._create_stale_check_file()
        elif self.args.pull and not (
                self.args.pull_daily and self._pulled_today()):
            self.pull()
            self._create_stale_check_file()

    @property
    def volumes(self):
        """
        Returns all volumes tied to the image's container configuration
        """
        config = self._image_metadata()['Config']
        if config.get('Volumes') is None:
            return []
        return list(config['Volumes'].keys())

    @property
    def entrypoint(self):
        return self._image_metadata()['Config']['Entrypoint']

    @property
    def user(self):
        return self._image_metadata()['Config']['User']

    @property
    def home(self):
        return path.join('/home', self.user)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slipway', 'slipway.client', 'slipway.xdg_open']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3,<6.0', 'docker>=4.0.0']

entry_points = \
{'console_scripts': ['slipway = slipway:main']}

setup_kwargs = {
    'name': 'slipway',
    'version': '0.11.0',
    'description': 'CLI tool for managing development containers',
    'long_description': None,
    'author': 'AGhost-7',
    'author_email': 'jonathan.boudreau.92@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

"""
Entrypoint script
"""

import os
from os import environ, path, system
import sys
import pwd


entrypoint = environ.get('SLIPWAY_ENTRYPOINT')
user_name = environ.get('SLIPWAY_USER')
docker_gid = environ.get('SLIPWAY_DOCKER_GID')

# All we need to do is correct the permisions in the volumes and call it a day.
user = pwd.getpwnam(user_name)
uid = user.pw_uid
gid = user.pw_gid

for volume in environ.get('SLIPWAY_VOLUMES').split(','):
    if len(volume) > 0:
        os.chown(volume, uid, gid)

# Also need to fix the permissions for the socket mount used on the host. This
# is created because we're bind mounting the xdg-open socket server and gpg.
os.chown('/run/user/{}'.format(uid), uid, gid)

if environ.get('SLIPWAY_DOCKER_GID'):
    system('groupadd --gid $SLIPWAY_DOCKER_GID slipway_docker')
    system('usermod -aG slipway_docker $SLIPWAY_USER')

args = ['sudo', '--preserve-env', '-u', user_name]

if entrypoint:
    args.append(entrypoint)

args.extend(sys.argv[1:])

env = environ.copy()
env['HOME'] = path.join('/home', user_name)

os.execvpe('sudo', args, env)
